#version 330 core

//===========================Defines
#define NR_POINT_LIGHTS 1
#define GODRAY_SAMPLES 100

//===========================SamplerDisks
vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1, 1), vec3(1, -1, 1), vec3(-1, -1, 1), vec3(-1, 1, 1), 
   vec3(1, 1, -1), vec3(1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1, 0), vec3(1, -1, 0), vec3(-1, -1, 0), vec3(-1, 1, 0),
   vec3(1, 0, 1), vec3(-1, 0, 1), vec3(1, 0, -1), vec3(-1, 0, -1),
   vec3(0, 1, 1), vec3(0, -1, 1), vec3(0, -1, -1), vec3(0, 1, -1)
);
float weight[5] = float[] (0.227027, 0.1945946, 0.1216216, 0.054054, 0.016216);

//===========================Input
in VS_OUT {
    vec3 Normal;
    vec3 FragPos;
    vec2 TexCoords;
    vec3 Tangent;
    vec3 Bitangent;
    mat3 TBN;
    vec3 TangentLightPos[NR_POINT_LIGHTS];
    vec3 TangentViewPos;
    vec3 TangentFragPos;
} fs_in;

//===========================Output
out vec4 color;

//===========================Uniforms
struct Screen {
	float width;
	float height;
};
uniform Screen screen;

struct Material {
    sampler2D diffuse;
	sampler2D specular;
	sampler2D normal;
    float shininess;
}; 
uniform Material material;

struct PointLight {
    vec3 position;
    
    float constant;
    float linear;
    float quadratic;
	
    vec3 ambient;
    vec3 diffuse;
	vec3 specular;
};
uniform PointLight pointLights[NR_POINT_LIGHTS];

struct GodRays {
	bool on;
    float exposure;
    float decay;
    float density;
    float weight;
    vec2 lightScreenPosition;
	sampler2D map;
};
uniform GodRays godRays;

struct Shadow {
	bool on;
	float far;
	samplerCube map;
};
uniform Shadow shadow;

struct Bloom {
	bool on;
	sampler2D map;
};
uniform Bloom bloom;

uniform vec3 viewPos;

//===========================Functions
float CalcShadow(vec3 fragPos, vec3 lightPos);
vec3 CalcBloom();
vec3 CalcPointLight(int i, PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir, vec3 bloom);
vec4 CalcGodray();

//===========================Main
void main()
{    
	vec3 norm = vec3(texture(material.normal, fs_in.TexCoords));
	norm = normalize(norm * 2.0 - 1.0);

    vec3 viewDir = normalize(fs_in.TangentViewPos - fs_in.TangentFragPos);

	vec3 result = vec3(0.0f);
	vec3 bloomValue = CalcBloom();

    for(int i = 0; i < NR_POINT_LIGHTS; i++)
        result += CalcPointLight(i, pointLights[i], norm, fs_in.FragPos, viewDir, bloomValue);

	color = vec4(result, 1.0);

	if(godRays.on) {
		color *= CalcGodray();
	}
	if(bloom.on) {
		color += vec4(bloomValue, 0.0f);
	}
}

float CalcShadow(vec3 fragPos, vec3 lightPos)
{
    vec3 fragToLight = fragPos - lightPos;
    
    float currentDepth = length(fragToLight);
    
    float shadowValue = 0.0;
    float bias = 0.02;
    int samples = 20;
    float viewDistance = length(viewPos - fragPos);
    float diskRadius = (1.0 + (viewDistance / shadow.far)) / 25.0;
    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(shadow.map, fragToLight + gridSamplingDisk[i] * diskRadius).r;
        closestDepth *= shadow.far;   // Undo mapping [0;1]
        if(currentDepth - bias > closestDepth)
            shadowValue += 1.0;
    }
    shadowValue /= float(samples);
        
    return shadowValue;
}

vec3 CalcBloom()
{
	vec2 tex_offset = 1.0 / textureSize(bloom.map, 0); // gets size of single texel
    vec3 result = texture(bloom.map, gl_FragCoord.xy/vec2(screen.width,screen.height)).rgb * weight[0];
    for(int i = 1; i < 5; ++i)
	{
		result += texture(bloom.map, gl_FragCoord.xy/vec2(screen.width,screen.height) + vec2(0.0, tex_offset.y * i)).rgb * weight[i];
        result += texture(bloom.map, gl_FragCoord.xy/vec2(screen.width,screen.height) - vec2(0.0, tex_offset.y * i)).rgb * weight[i];
	}
    return result;
}

vec3 CalcPointLight(int i, PointLight light, vec3 normal, vec3 fragPos, vec3 viewDir, vec3 bloomValue)
{
    vec3 lightDir = normalize(fs_in.TangentLightPos[i] - fs_in.TangentFragPos);

    float diff = max(dot(normal, lightDir), 0.0);
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0), material.shininess);
    float distance = length(light.position - fragPos);
    float attenuation = 1.0f / (light.constant + light.linear * distance + light.quadratic * (distance * distance));

    vec3 ambient = light.ambient * vec3(texture(material.diffuse, fs_in.TexCoords)) * attenuation;
    vec3 diffuse = light.diffuse * diff * vec3(texture(material.diffuse, fs_in.TexCoords)) * attenuation;
    vec3 specular = light.specular * spec * vec3(texture(material.specular, fs_in.TexCoords)) * attenuation;

    float shadow = shadow.on ? CalcShadow(fs_in.FragPos, light.position) : 0.0;  

	// Bloom illumination fix for additive blending
	if(bloomValue.r < 0.17f && bloomValue.g < 0.17f && bloomValue.b < 0.17f || !bloom.on) {
		return (ambient + (1.0f - shadow) * (diffuse + specular));
	} else {
		return (ambient + (1.0f - shadow) * (diffuse + specular) * bloomValue);
	}
}

vec4 CalcGodray()
{
	vec4 godrayColor = vec4(0.0f,0.0f,0.0f,1.0f);

	vec2 deltaTextCoord = vec2( (gl_FragCoord.xy/vec2(screen.width,screen.height)*2.0f - vec2(1.0f,1.0f)) - godRays.lightScreenPosition);
	vec2 textCoo = gl_FragCoord.xy/vec2(screen.width,screen.height)*2.0f - vec2(1.0f,1.0f);
	deltaTextCoord *= 1.0 /  float(GODRAY_SAMPLES) * godRays.density;
	float illuminationDecay = 1.0;

	for(int i=0; i < GODRAY_SAMPLES; i++)
	{
		textCoo -= deltaTextCoord;
		vec4 sample = texture2D(godRays.map, (textCoo + vec2(1.0f, 1.0f))/2.0f);

		sample *= illuminationDecay * godRays.weight;

		godrayColor += sample;
		
		illuminationDecay *= godRays.decay;
	}

	return godrayColor * godRays.exposure;
}